var _geo_fence_callback = null
var _geo_fence_object = []

/**
 * _geo_fence_object
 * 
 *  Elements ['name']
 *  {
 *      name:
 *      isInside:
 *      type:
 * 
 *      // for polygon
 *      path:
 *      
 *      // for circle
 *      lat:
 *      lng:
 *      radius:
 *  }
 * 
 */

app.CreateGeoFence = function() {
    return new _Geo_Fence()
}
function _Geo_Fence() {
    // functions definitions
    function _DegreeToRadians(deg) {
        return deg * Math.PI / 180;
    }
    function _GetDistance(fence, lat, lng) {
        
        // a = sin²(Δφ/2) + cos(φ1)⋅cos(φ2)⋅sin²(Δλ/2)
        // δ = 2·atan2(√(a), √(1−a))
        // see mathforum.org/library/drmath/view/51879.html for derivation
    
        const R = 6371e3
        const φ1 = fence.lat,  λ1 = fence.lng
        const φ2 = _DegreeToRadians(lat), λ2 = _DegreeToRadians(lng)
        const Δφ = φ2 - φ1
        const Δλ = λ2 - λ1
    
        const a = Math.sin(Δφ/2)*Math.sin(Δφ/2) + Math.cos(φ1)*Math.cos(φ2) * Math.sin(Δλ/2)*Math.sin(Δλ/2)
        const c = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1-a))
        const d = R * c
        return d
    }
    function _CheckIfLocationIsInside(poly, pt) {
        for(var c = false, i = -1, l = poly.length, j = l - 1; ++i < l; j = i)
            ((poly[i].lng <= pt.lng && pt.lng < poly[j].lng) || (poly[j].lng <= pt.lng && pt.lng < poly[i].lng))
            && (pt.lat < (poly[j].lat - poly[i].lat) * (pt.lng - poly[i].lng) / (poly[j].lng - poly[i].lng) + poly[i].lat)
            && (c = !c);
        return c;
    }
    
    //Methods definitions
    this.AddPolygon = function(path, name) {
        if(!path) throw new Error("Missing path for a new polygon")
        else if(_geo_fence_object.length>1&&!name) throw new Error("Name is required for more than one geofence")
        var newPoly = {
            name: name,
            type: "polygon",
            path: path,
            isInside: false
        }
        if(!name) _geo_fence_object.push(newPoly)
        else _geo_fence_object[name] = newPoly
    }
    this.SetPolygon = function(path, name) {
        if(!path) throw new Error("Missing path for the polygon")
        else if(_geo_fence_object.length>1&&!name) throw new Error("Name is required")
        if(_geo_fence_object.length == 1) {
            if(_geo_fence_object[0] && _geo_fence_object[0].type == "polygon") _geo_fence_object[0].path = path
            else if(_geo_fence_object[name].type == "polygon") _geo_fence_object[name].path = path
            else throw new Error("Defined geofence is not a polygon")
        } else {
            if(_geo_fence_object[name].type == "polygon") _geo_fence_object[name].path = path
            else throw new Error("Defined geofence is not a polygon")
        }
    }
    
    this.AddCircle = function(lat, lng, radius, name) {
        if(!lat || !lng || !radius) throw new Error("Missing argument for new circle")
        else if(_geo_fence_object.length>1&&!name) throw new Error("Name is required for more than one geofence")
        var newCircle = {
            name: name,
            type: "circle",
            lat: lat,
            lng: lng,
            radius: radius,
            isInside: false
        }
        if(!name) _geo_fence_object.push(newCircle)
        else _geo_fence_object[name] = newCircle
    }
    this.SetCircle = function(lat, lng, radius, name) {
        if(!lat || !lng || !radius) throw new Error("Missing argument for new circle")
        else if(_geo_fence_object.length>1&&!name) throw new Error("Name is required")
        if(_geo_fence_object.length == 1) {
            if(_geo_fence_object[0] && _geo_fence_object[0].type == "circle") {
                _geo_fence_object[0].lat = lat
                _geo_fence_object[0].lng = lng
                _geo_fence_object[0].radius = radius
            }
            else if(_geo_fence_object[name].type == "circle") {
                _geo_fence_object[name].lat = lat
                _geo_fence_object[name].lng = lng
                _geo_fence_object[name].radius = radius
            } else throw new Error("Defined geofence is not a circle")
        } else {
            if(_geo_fence_object[name].type == "circle") {
                _geo_fence_object[name].lat = lat
                _geo_fence_object[name].lng = lng
                _geo_fence_object[name].radius = radius    
            }
            else throw new Error("Defined geofence is not a circle")   
        }
    }
    this.SetCenter = function(lat, lng, name) {
        if(!lat || !lng) throw new Error("Missing argument for new circle")
        else if(_geo_fence_object.length>1&&!name) throw new Error("Name is required")
        if(_geo_fence_object.length == 1) {
            if(_geo_fence_object[0] && _geo_fence_object[0].type == "circle") {
                _geo_fence_object[0].lat = lat
                _geo_fence_object[0].lng = lng
            }
            else if(_geo_fence_object[name].type == "circle") {
                _geo_fence_object[name].lat = lat
                _geo_fence_object[name].lng = lng
            } else throw new Error("Defined geofence is not a circle")
        } else {
            if(_geo_fence_object[name].type == "circle") {
                _geo_fence_object[name].lat = lat
                _geo_fence_object[name].lng = lng
            } else throw new Error("Defined geofence is not a circle")   
        }
    }
    this.SetRadius = function(radius, name) {
        if(!radius) throw new Error("Radius is required for SetRadius")
        else if(_geo_fence_object.length>1&&!name) throw new Error("Name is required")
        if(_geo_fence_object.length == 1) {
            if(_geo_fence_object[0] && _geo_fence_object[0].type == "circle") _geo_fence_object[0].radius = radius
            else if(_geo_fence_object[name].type == "circle") _geo_fence_object[name].radius = radius
            else throw new Error("Defined geofence is not a circle")
        } else {
            if(_geo_fence_object[name].type == "circle") _geo_fence_object[name].radius = radius
            else throw new Error("Defined geofence is not a circle")   
        }
    }
    this.SetOnChange = function(callback) {
        _geo_fence_callback = callback
    }
    this.IsInside = function(name) {
        var returnValue = null
        if(name && _geo_fence_object[name] && _geo_fence_object.length == 1) {
            returnValue = _geo_fence_object[name].isInside
        } else {
            returnValue = []
            _geo_fence_object.forEach(function(item) {
                returnValue[item.name] = item.isInside
            })
        }
        return returnValue
    }
    
    //locator object
    this._locator = app.CreateLocator("GPS,Network")
    this._locator.SetOnChange(function(pos) {
        _geo_fence_object.forEach(function(fence) {
            if(fence.type == "circle") {
                var s = _GetDistance(fence, pos.latitude, pos.longitude)
                var is_inside = fence.isInside
                fence.isInside = s<fence.radius?true:false
                if(fence.isInside !== is_inside && _geo_fence_callback) return _geo_fence_callback(fence.isInside, fence.name)
            } else {
                var location = {lat: pos.latitude, lng: pos.longitude}
                var is_inside = fence.isInside
                fence.isInside = _CheckIfLocationIsInside(fence.path, location)
                if(fence.isInside !== is_inside && _geo_fence_callback) return _geo_fence_callback(fence.isInside, fence.name)
            }
        })
        
    })
    this.Start = function() {
        this._locator.Start()
    }
    this.Stop = function(name) {
        this._locator.Stop()
    }
    this.ClearWatch = function(name) {
        if(name) _geo_fence_object[name] = null
        else {
            _geo_fence_object = []
            this._locator.Stop()
        }
    }
}

